from . import generic

arch = 'MIPS64'

class R_MIPS_64(generic.GenericAbsoluteAddendReloc):
    pass

class R_MIPS_REL32(generic.GenericRelativeReloc):
    pass

class R_MIPS_COPY(generic.GenericCopyReloc):
    pass

class R_MIPS_TLS_DTPMOD64(generic.GenericTLSModIdReloc):
    pass

class R_MIPS_TLS_DTPREL64(generic.GenericTLSDoffsetReloc):
    pass

class R_MIPS_TLS_TPREL64(generic.GenericTLSOffsetReloc):
    pass



import re
import logging
import binascii
import struct

from . import register_backend, Backend
from ..errors import CLEError

l = logging.getLogger(name=__name__)

__all__ = ('Hex',)

intel_hex_re = re.compile(b":([0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])"
                          b"([0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F]+)*([0-9a-fA-F][0-9a-fA-F])")

HEX_TYPE_DATA = 0x00
HEX_TYPE_EOF = 0x01
HEX_TYPE_EXTSEGADDR = 0x02
HEX_TYPE_STARTSEGADDR = 0x03
HEX_TYPE_EXTLINEARADDR = 0x04
HEX_TYPE_STARTLINEARADDR = 0x05

if bytes is not str:
    chh = lambda x: x
else:
    chh = ord


class Hex(Backend):
    """
    A loader for Intel Hex Objects
    See https://en.wikipedia.org/wiki/Intel_HEX
    """
    is_default = True # Tell CLE to automatically consider using the Hex backend

    @staticmethod
    def parse_record(line):
        m = intel_hex_re.match(line)
        if not m:
            raise CLEError("Invalid HEX record: " + line)
        my_cksum = 0
        count, addr, rectype, data, cksum = m.groups()
        cksum = int(cksum, 16)
        for d in binascii.unhexlify(line[1:-2]):
            my_cksum = (my_cksum + chh(d)) % 256
        my_cksum = ((my_cksum ^ 0xff) + 1) % 256
        if my_cksum != cksum:
            raise CLEError("Invalid checksum: Computed %s, found %s" % (hex(my_cksum), hex(cksum)))
        count = int(count, 16)
        addr = int(addr, 16)
        rectype = int(rectype, 16)
        if data:
            data = binascii.unhexlify(data)
        if data and count != len(data):
            raise CLEError("Data length field does not match length of actual data: " + line)
        return rectype, addr, data

    @staticmethod
    def coalesce_regions(regions):
        # Lots of tiny memory regions is bad!
        # The greedy algorithm to smash them together:
        result = []
        for addr, region in sorted(regions):
            if result and result[-1][0] + len(result[-1][1]) == addr:
                result[-1] = (result[-1][0], result[-1][1] + region)
            else:
                result.append((addr, region))

        return result

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.arch is None:
            raise CLEError("To use the Hex binary backend, you need to specify an architecture in the loader options.")

        # Do the whole thing in one shot.
        self.os = 'unknown'
        got_base = False
        got_entry = False
        self._binary_stream.seek(0)
        string = self._binary_stream.read()
        recs = string.splitlines()
        regions = []
        max_addr = 0
        min_addr = 0xffffffffffffffff
        self._base_address = 0
        for rec in recs:
            rectype, addr, data = Hex.parse_record(rec)
            if rectype == HEX_TYPE_DATA:
                addr += self._base_address
                #l.debug("Loading %d bytes at " % len(data) + hex(addr))
                # Raw data.  Put the bytes
                regions.append((addr, data))
                # We have to be careful about the min and max addrs
                if addr < min_addr:
                    min_addr = addr
                max_addr = max(max_addr, addr + len(data) - 1)
            elif rectype == HEX_TYPE_EOF:
                # EOF
                l.debug("Got EOF record.")
                break
            elif rectype == HEX_TYPE_EXTSEGADDR:
                # "Extended Mode" Segment address, take this value, multiply by 16, make the base
                self._base_address = struct.unpack('>H', data)[0] * 16
                got_base = True
                l.debug("Loading a segment at %#x", self._base_address)
            elif rectype == HEX_TYPE_STARTSEGADDR:
                # Four bytes, the segment and the initial IP
                got_base = True
                got_entry = True
                self._initial_cs, self._initial_ip = struct.unpack('>HH', data)
                # The whole thing is the entry, as far as angr is concerned.
                self._entry = struct.unpack('>I', data)[0]
                l.debug("Got entry point at %#x", self._entry)
            elif rectype == HEX_TYPE_EXTLINEARADDR:
                got_base = True
                # Specifies the base for all future data bytes.
                self._base_address = struct.unpack('>H', data)[0] << 16
                l.debug("Loading a segment at %#x", self._base_address)
            elif rectype == HEX_TYPE_STARTLINEARADDR:
                got_entry = True
                # The 32-bit EIP, really the same as STARTSEGADDR, but some compilers pick one over the other.
                self._entry = struct.unpack('>I', data)[0]
                l.debug("Found entry point at %#x", self._entry)
                self._initial_eip = self._entry
            else:
                raise CLEError("This HEX Object type is not implemented: " + hex(rectype))
        if not got_base:
            l.warning("No base address was found in this HEX object file. It is assumed to be 0")
        if not got_entry:
            l.warning("No entry point was found in this HEX object file, and it is assumed to be 0. "
                      "Specify one with `entry_point` to override.")
        # HEX specifies a ton of tiny little memory regions.  We now smash them together to make things faster.
        new_regions = Hex.coalesce_regions(regions)
        for addr, data in new_regions:
            self.memory.add_backer(addr, data)
        self._max_addr = max_addr
        self._min_addr = min_addr

    @staticmethod
    def is_compatible(stream):
        stream.seek(0)
        s = stream.read(0x10)
        stream.seek(0)
        return s.startswith(b":")

register_backend("hex", Hex)



import os.path

from typing import Dict, List
from .variable import Variable
from .subprogram import Subprogram


class CompilationUnit:
    """
    CompilationUnit for DWARF
    See http://dwarfstd.org/doc/DWARF5.pdf page 60
    """
    functions: Dict[int, Subprogram] = {}
    global_variables: List[Variable] = []

    def __init__(self, name, comp_dir, low_pc, high_pc, language):
        self.name = name
        self.comp_dir = comp_dir
        self.file_path = os.path.join(self.comp_dir, self.name)
        self.low_pc = low_pc
        self.high_pc = high_pc
        self.language = language



import os
import logging

from .backends.elf.metaelf import MetaELF
from .errors import CLEFileNotFoundError

l = logging.getLogger(name=__name__)

def convert_info_sharedlibrary(fname):
    """
    Convert a dump from gdb's ``info sharedlibrary`` command to a set of options that can be passed to CLE to replicate
    the address space from the gdb session

    :param fname:   The name of a file containing the dump
    :returns:       A dict appropriate to be passed as ``**kwargs`` for ``angr.Project`` or ``cle.Loader``
    """
    return _parse_gdb_map(fname, True)

def convert_info_proc_maps(fname):
    """
    Convert a dump from gdb's ``info proc maps`` command to a set of options that can be passed to CLE to replicate
    the address space from the gdb session

    :param fname:   The name of a file containing the dump
    :returns:       A dict appropriate to be passed as ``**kwargs`` for ``angr.Project`` or ``cle.Loader``
    """
    return _parse_gdb_map(fname, False)

def _parse_gdb_map(gdb_map, gdb_fix):
    if not os.path.isfile(gdb_map):
        raise CLEFileNotFoundError("gdb mapping file %s does not exist" % gdb_map)
    with open(gdb_map, 'r') as f:
        data = f.readlines()

    gmap = {}
    for line in data:
        if line in ('\n', '\r\n'):
            continue
        line_items = line.split()
        # Get rid of all metadata, just extract lines containing addresses
        if "0x" not in line_items[0]:
            continue
        if line_items[-1].startswith('['):
            continue
        try:
            int(line_items[-1], 16)
        except ValueError:
            pass
        else:
            continue

        addr, objfile = int(line_items[0], 16), line_items[-1].strip()

        # Get the smallest address of each libs' mappings
        try:
            gmap[objfile] = min(gmap[objfile], addr)
        except KeyError:
            gmap[objfile] = addr

    # Find lib names
    #libnames = filter(lambda n: '.so' in n, gmap.keys())

    # Find base addr for each lib (each lib is mapped to several segments,
    # we take the segment that is loaded at the smallest address).
    lib_opts = {}
    main_opts = {}
    force_load_libs = []
    smallest_addr = min(gmap.values())

    for lib, addr in gmap.items():
        if addr == smallest_addr and not gdb_fix:
            # this is the main binary
            opts = main_opts
        else:
            #if not os.path.exists(lib):
            #    lib = _simple_search(lib)
            force_load_libs.append(lib)
            opts = {}
            lib_opts[lib] = opts

        # address of .text -> base address of the library
        if gdb_fix:
            found = _simple_search(lib)
            if found:
                addr = addr - MetaELF.get_text_offset(found)

        l.info("gdb_plugin: mapped %s to %#x", lib, addr)
        opts["base_addr"] = addr

    return {'force_load_libs': force_load_libs, 'main_opts': main_opts, 'lib_opts': lib_opts}

GDB_SEARCH_PATH = ['/lib', '/usr/lib']

def _simple_search(libname):
    dirs = list(GDB_SEARCH_PATH)
    while dirs:
        dirname = dirs.pop(0)
        try:
            for name in os.listdir(dirname):
                if name in ('.', '..'):
                    continue
                full = os.path.join(dirname, name)
                if os.path.isdir(full):
                    if full.count('/') < 12: # don't go too deep
                        dirs.append(full)
                if os.path.isfile(full) and name == libname:
                    return full
        except OSError:
            pass
    return libname



import struct
import logging
from .pereloc import PEReloc
from ....address_translator import AT

l = logging.getLogger(name=__name__)

class DllImport(PEReloc):
    """
    There's nothing special to be done for DLL imports but this class
    provides a unique name to the relocation type.
    """
    pass

class IMAGE_REL_BASED_ABSOLUTE(PEReloc):
    def relocate(self):
        pass

class IMAGE_REL_BASED_HIGHADJ(PEReloc):
    def __init__(self, owner, addr, next_rva):
        super().__init__(owner, None, addr)
        self.next_rva = next_rva
    @property
    def value(self):
        """
        In all the other cases, we can ignore the relocation difference part of the
        calculation because we simply use to_mva() to get our rebased address. In this
        case, however, we have to adjust the un-rebased address first.
        """
        org_bytes = self.owner.memory.load(self.relative_addr, 2)
        org_value = struct.unpack('<I', org_bytes)[0]
        adjusted_value = (org_value << 16) + self.next_rva
        adjusted_value = (AT.from_lva(adjusted_value, self.owner) & 0xffff0000) >> 16
        adjusted_bytes = struct.pack('<I', adjusted_value)
        return adjusted_bytes

class IMAGE_REL_BASED_HIGHLOW(PEReloc):
    @property
    def value(self):
        org_bytes = self.owner.memory.load(self.relative_addr, 4)
        org_value = struct.unpack('<I', org_bytes)[0]
        rebased_value = AT.from_lva(org_value, self.owner).to_mva()
        rebased_bytes = struct.pack('<I', rebased_value & 0xffffffff)
        return rebased_bytes

class IMAGE_REL_BASED_DIR64(PEReloc):
    @property
    def value(self):
        org_bytes = self.owner.memory.load(self.relative_addr, 8)
        org_value = struct.unpack('<Q', org_bytes)[0]
        rebased_value = AT.from_lva(org_value, self.owner).to_mva()
        if rebased_value < 0 or rebased_value >= 0x10000000000000000:
            l.error("Incorrect rebased address %x found at %s. Maybe the relocation table is broken.",
                    rebased_value, self.owner)
            return None
        rebased_bytes = struct.pack('<Q', rebased_value)
        return rebased_bytes

class IMAGE_REL_BASED_HIGH(PEReloc):
    @property
    def value(self):
        org_bytes = self.owner.memory.load(self.relative_addr, 2)
        org_value = struct.unpack('<H', org_bytes)[0]
        rebased_value = AT.from_lva(org_value, self.owner).to_mva()
        adjusted_value = (rebased_value >> 16) & 0xffff
        adjusted_bytes = struct.pack('<H', adjusted_value)
        return adjusted_bytes

class IMAGE_REL_BASED_LOW(PEReloc):
    @property
    def value(self):
        org_bytes = self.owner.memory.load(self.relative_addr, 2)
        org_value = struct.unpack('<H', org_bytes)[0]
        rebased_value = AT.from_lva(org_value, self.owner).to_mva()
        adjusted_value = rebased_value & 0x0000FFFF
        adjusted_bytes = struct.pack('<H', adjusted_value)
        return adjusted_bytes
